#ifndef SIKE_R1_NAMESPACE_H
#define SIKE_R1_NAMESPACE_H

#define random_mod_order_B random_mod_order_B_r1
#define EphemeralKeyGeneration_A EphemeralKeyGeneration_A_r1
#define EphemeralKeyGeneration_B EphemeralKeyGeneration_B_r1
#define EphemeralSecretAgreement_A EphemeralSecretAgreement_A_r1
#define EphemeralSecretAgreement_B EphemeralSecretAgreement_B_r1
#define xDBL xDBL_r1
#define xDBLe xDBLe_r1
#define get_4_isog get_4_isog_r1
#define eval_4_isog eval_4_isog_r1
#define xTPL xTPL_r1
#define xTPLe xTPLe_r1
#define get_3_isog get_3_isog_r1
#define eval_3_isog eval_3_isog_r1
#define inv_3_way inv_3_way_r1
#define get_A get_A_r1
#define j_inv j_inv_r1
#define xDBLADD xDBLADD_r1
#define swap_points swap_points_r1
#define LADDER3PT LADDER3PT_r1
#define load64 load64_r1
#define store64 store64_r1
#define KeccakF1600_StatePermute KeccakF1600_StatePermute_r1
#define keccak_absorb keccak_absorb_r1
#define keccak_squeezeblocks keccak_squeezeblocks_r1
#define cshake256_simple_absorb cshake256_simple_absorb_r1
#define cshake256_simple cshake256_simple_r1
#define digit_x_digit digit_x_digit_r1
#define mp_mul mp_mul_r1
#define rdc_mont rdc_mont_r1
#define to_mont to_mont_r1
#define from_mont from_mont_r1
#define copy_words copy_words_r1
#define mp_addfast mp_addfast_r1
#define mp_addfastx2 mp_addfastx2_r1
#define mp_sub mp_sub_r1
#define mp_subfast mp_subfast_r1
#define to_fp2mont to_fp2mont_r1
#define from_fp2mont from_fp2mont_r1
#define mp_add mp_add_r1
#define mp_shiftr1 mp_shiftr1_r1
#define Alice_order Alice_order_r1
#define Bob_order Bob_order_r1
#define A_gen A_gen_r1
#define B_gen B_gen_r1
#define Montgomery_R2 Montgomery_R2_r1
#define Montgomery_one Montgomery_one_r1
#define Montgomery_Rprime Montgomery_Rprime_r1
#define Montgomery_rprime Montgomery_rprime_r1
#define Border_div3 Border_div3_r1
#define strat_Alice strat_Alice_r1
#define strat_Bob strat_Bob_r1
#define clear_words clear_words_r1
#define init_basis init_basis_r1
#define fp2_encode fp2_encode_r1
#define fp2_decode fp2_decode_r1

#endif // SIKE_R1_NAMESPACE_H
